/******************************************************************************
** Az albbi program s informcik a Nemzeti Kutatsi Fejlesztsi Programok
** 2002. vi plyzatn a 2. sz. programban elfogadott NKFP-2/0017/2002 szm
** ,,Adatrosta - webtrhzak nagytmeg adatbzisainak elemzse adatbnyszati 
** s statisztikai eszkzk segtsgvel'' elnevezs projekt termke.
**
** A program a projekt folyamn feldolgozsra kerl naplllomnyok
** (web, access s mail) binris kdolst s tmrtst vgzi.
**
** A program s informcik az Adatrosta projekt konzorciumnak tulajdona.
** Engedly nlkli felhasznlsa tilos.
**
** Az Adatrosta projekt konzorcium tagjai:
** ELTE TTK Szmtgptudomnyi Tsz.
** BME Matematika Intzet
** MTA SZTAKI Informatikai Kutatlaboratrium
** Axelero Rt.
** econet.hu Rt.
**
** http://adatrosta.hu/
** email: info@adatrosta.hu
**
** (C) 2003. Minden jog fenntartva.
**
**
** This data and information are property of the "Data Riddle" consortium.
** Modification or use without permission is strictly prohibited.
**
** (C) 2003 All rights reserved.
**
** Author: RACZ Balazs, 2003.
**
******************************************************************************/

/*!
  \file   configurator.cpp
  \brief  implementation of configurator class

  \author RACZ Balazs
  \date   2003-01-19
*/


#include <stdio.h>
#include "configurator.h"
#include "log.h"
#include <algorithm> //for min() and max()
#include <string>

using namespace std;

configuration::configuration() {
  inEnum = false;
  subkeys.clear();
  values.clear();
  path = "";
  lastused = -1;
}

configuration::~configuration() {
  for (subkeys_t::iterator i=subkeys.begin();i!=subkeys.end();i++) {
    delete i->second;
  }
}

string configuration::getPath() {
  return path;
}

void configuration::startEnum() {
  inEnum = true;
  enumPos = 0;
}

bool configuration::endEnum() {
  if (!inEnum)
    throw EEnumNotStarted();
  return (enumPos>=(int)values.size());
}

void configuration::stepEnum() {
  if (!inEnum)
    throw EEnumNotStarted();
  enumPos++;
}

string configuration::getValue_str() {
  if (!inEnum)
    throw EEnumNotStarted();
  if (endEnum())
    throw EEndOfEnumeration();
  lastused = max(lastused,enumPos);
  return values[enumPos];
}

int configuration::getValue_int(){
  string v=getValue_str(); //this will throw the necessary exceptions
  int i=atoi(v.c_str());
  return i;
}

double configuration::getValue_double(){
  string v=getValue_str(); //this will throw the necessary exceptions
  double d=atof(v.c_str());
  return d;
}

int configuration::getValue_range(int low, int high) {
  int i=getValue_int(); //this will throw the necessary exceptions
  int ret=i;
  if ((i<low)||(i>high)) {
    ret=min(max(i,low),high);
    err(1,"parameter %s out of bounds: %s=%d not in [%d,%d]; clipping to %d",path.c_str(),values[enumPos].c_str(),i,low,high,ret);
    return ret; 
  } else 
    return i;
}


string configuration::trim(const string & s) {
  string::size_type begin_data = s.find_first_not_of(whitespace);
  if (begin_data==string::npos)
    return "";
  string::size_type end_data = s.find_last_not_of(whitespace);
  return s.substr(begin_data,end_data-begin_data+1);
}

const char * configuration::whitespace = "\n\r 	";

void configuration::appendFromFile(istream & f) {
  const int buflen = 1000;
  char buf[buflen];
  string s;
  string prevKey="";
  while (f.good()) {
    f.getline(buf,buflen);
    //    dbg( 3, "line: \"%s\"", buf);
    s=buf;
    s=trim(s);
    //    cerr << s;
    //look for comment
    if ((!s.size()) || (s[0]=='#'))
      continue;
    //find equal sign
    string::size_type pos_eq=s.find_first_of('=');
    if (pos_eq==string::npos) {
      //No = sign in input file
      warn( 1,"ignoring line without equal sign in configuration: %s",buf);
      continue;
    }
    //break into key and value
    string key = trim(s.substr(0,pos_eq));
    //    dbg( 3, "key= \"%s\"",key.c_str());
    string value = trim(s.substr(pos_eq+1));
    //handle quoting
    if ((value[0]=='"')&&(value[value.size()-1]=='"')) {
      value = value.substr(1,value.size()-2);
    }
    //handle previous Key continuation
    if (key[0]=='.') 
      key=prevKey+key;
    string::size_type pos_dot;
    // fill prevKey for the next iteration
    pos_dot = key.find_last_of('.');
    if (pos_dot==string::npos) {
      //no dots in the key
      prevKey="";
    } else {
      prevKey=key.substr(0,pos_dot); //this doesn't include the '.' character
    }
    //traverse configuration tree
    configuration *current = this;
    string newpath = path;
    while (key!="") {
      pos_dot=key.find_first_of('.');
      string firstpart=trim(key.substr(0,pos_dot));
      if (pos_dot==string::npos) 
	key="";
      else
	key=trim(key.substr(pos_dot+1));
      newpath.append(".");
      newpath.append(firstpart);
      configuration *next = current->getKey(firstpart);
      if (!next) {
	next = new configuration();
	next->path = newpath;
	current->subkeys.push_back(pair<string,configuration*>(firstpart,next));
      } else {
	if(newpath!=next->path) {
	  warn( 0, "bug! inconsistent tree: path %s != stored path %s",newpath.c_str(),next->path.c_str());
	  newpath=next->path;
	}
      }
      current=next;
    } //while traverse configuration tree
    //now: we have a configuration and a value
    current->values.push_back(value);
  } //while not eof(fin)
  if (f.fail()) {
    warn( 1, "minor error reading input stream. Possibly no end of line in the last line?");    
  }
  if (f.bad()) {
    warn( 0, "major error reading input stream.");        
  }
}

void configuration::warnIfUnused() {
  //My own unused variables:
  for (unsigned int i=lastused+1;i<values.size();i++) {
    warn( 1, "unused config option: %s = \"%s\"", path.c_str(), values[i].c_str());
  }
  //Recursively process subkeys
  for(subkeys_t::iterator i=subkeys.begin();i!=subkeys.end();i++) {
    i->second->warnIfUnused();
  }
}


configuration* configuration::getKey(string name) {
  subkeys_t::iterator i=subkeys.begin();
  while ((i!=subkeys.end())&&(i->first!=name)) 
    i++;
  //  map<string,configuration*>::iterator i=subkeys.find(name);
  if (i!=subkeys.end()) {
    return i->second;
  } else {
    return NULL;
  }
}

vector<pair<string,configuration*> > configuration::getSubKeys() {
  return vector<pair<string,configuration*> >(subkeys.begin(),subkeys.end());
}

