/******************************************************************************
** Az albbi program s informcik a Nemzeti Kutatsi Fejlesztsi Programok
** 2002. vi plyzatn a 2. sz. programban elfogadott NKFP-2/0017/2002 szm
** ,,Adatrosta - webtrhzak nagytmeg adatbzisainak elemzse adatbnyszati 
** s statisztikai eszkzk segtsgvel'' elnevezs projekt termke.
**
** A program a projekt folyamn feldolgozsra kerl naplllomnyok
** (web, access s mail) binris kdolst s tmrtst vgzi.
**
** A program s informcik az Adatrosta projekt konzorciumnak tulajdona.
** Engedly nlkli felhasznlsa tilos.
**
** Az Adatrosta projekt konzorcium tagjai:
** ELTE TTK Szmtgptudomnyi Tsz.
** BME Matematika Intzet
** MTA SZTAKI Informatikai Kutatlaboratrium
** Axelero Rt.
** econet.hu Rt.
**
** http://adatrosta.hu/
** email: info@adatrosta.hu
**
** (C) 2003. Minden jog fenntartva.
**
**
** This data and information are property of the "Data Riddle" consortium.
** Modification or use without permission is strictly prohibited.
**
** (C) 2003 All rights reserved.
**
** Author: RACZ Balazs, 2003.
**
******************************************************************************/

/*!
  \file   configurator.h
  \brief  class for reading configuration parameters

  Requires C++ STL multimap

  \author RACZ Balazs
  \date   2003-01-19
*/

#ifndef __CONFIGURATOR_H__
#define __CONFIGURATOR_H__

#include <map>
#include <string>
#include <exception>
#include <vector>

#include <iostream>

#include "types.h"

using namespace std;

//! \brief Exception thrown by class \c configuration when the enumeration has come to the end and more values are requested.
class EEndOfEnumeration: public exception {
};

//! \brief Exception thrown by class \c configuration when the enumeration has not yet started but values are requested. 
class EEnumNotStarted: public exception {
};

//! \brief holds configuration options (parameters) for an identity
/** 
   This class is a powerful tree based configuration interface.
   Subkeys are allowed and represented by other objects of this class.

   Currently values are strings, but can be retrieved as integers or
   integer-intervals.

   The class can read the configuration tree from a file.
   
   \todo subkey handling is ugly.
*/
class configuration {
public:

  //! \brief Position in the configuration tree
  /*!
    \return a string: '.'-separated path of ancestors
  */
  string getPath();

  //! \brief Starts an enumeration of the values associated with this node
  void startEnum();

  //! \brief Returns true, if there are no more values associated with this node
  /**
     \throws EEnumNotStarted if no previous call to startEnum() has been made
  */
  bool endEnum();

  //! \brief Moves enumeration iterator to next element.
  /** 
      \throws EEnumNotStarted if no previous call to startEnum() has been made
   */
  void stepEnum();

  //! \brief returns the value under the enumeration iterator 
  /** 	
	\throws EEndOfEnumeration if the enumeration iterator is behind the end
	\throws EEnumNotStarted if no previous call to startEnum() has been made
  */
  string getValue_str();
  
  //! \brief returns the value under the enumeration iterator converted to integer
  /**
     Displays an error message if the value cannot be converted to an integer
     \throws EEndOfEnumeration if the enumeration iterator is behind the end   
     \throws EEnumNotStarted if no previous call to startEnum() has been made  
  */
  int getValue_int();
  
  //! \brief returns the value under the enumeration iterator converted to double
  /**
     Displays an error message if the value cannot be converted to a double
     \throws EEndOfEnumeration if the enumeration iterator is behind the end   
     \throws EEnumNotStarted if no previous call to startEnum() has been made  
  */
  double getValue_double();
     
  //! \brief returns the value under the enumeration iterator converted to integer
  /**
     Displays an error message if the value cannot be converted to an integer
     or it is out of the specified range.
     \param low Lowest allowed return value
     \param high Highest allowed return value

     \throws EEndOfEnumeration if the enumeration iterator is behind the end  
     \throws EEnumNotStarted if no previous call to startEnum() has been made
  */
  int getValue_range(int low, int high);

  configuration();
  ~configuration();

  //! Adds parameters based on the content of a text stream
  /**
     Configuration file format:
     Empty lines and lines beginning with # are treated as comments
       and ignored
     Leading and trailing whitespaces (space, tab) are removed.
     Line is broken into key and value at the first '=' character
     Key and value is trimmed.
     If value starts and ends with '"' then they are removed. (Thus if
     you want to enter a value which ends in spaces or tabs, you have
       to enclose it in double quotes)
     Key is a '.'-separated list of subkey names (they are trimmed).
     If it starts with a '.' then it is relative to the previous key
       without its last part

     For an example, see origo.cfg:

\include origo.cfg
\verbatim
# basic objects
create.coord=ecoordinator
create.input=csvinput
create.fields=srcfieldcollection
create.filter=nonefilter
create.sorter=sorter

# basic parameters of coord
coord.sourcefilename=/home/bracz/adat/origo_frontend1_20021202.log
coord.destinationfilename=origo1_021202.dat
coord.blocksize=1000000
 #links of coord
coord.input=input
coord.filter=filter
coord.sorter=sorter
coord.fields=fields

#parmeters for the sorter
sorter.headerlen=1
sorter.bodylen=4
sorter.bodyoffset=100

#the source field collection: these start the pipes
# we create and register them to out srcFieldCollection
fields.create.srcusercookie=ifield_string
fields=srcusercookie
fields.create.srcdoc=ifield_string
fields=srcdoc
fields.create.srcdate=ifield_int64
fields=srcdate
fields.create.srcuseragent=ifield_string
fields=srcuseragent
fields.create.srccount=ifield_int32
fields=srccount

#the transcoding pipe
fields.create.codeuser=t_string_transcode
	.codedoc=t_string_transcode
	.codeagent=t_string_transcode

	.sortuser=sorterfield_int
	.sortdoc=sorterfield_int
	.sortdate=sorterfield_int64
	.sortcount=sorterfield_int
	.sortagent=sorterfield_int

	.outcount=32bitwriter
	.outuser=32bitwriter
	.outdoc=32bitwriter
	.outdate=64bitwriter
	.outagent=32bitwriter

#parameters for the source fields
fields.srcusercookie.link=codeuser
	.inputfieldname=cookie
	.fieldrole='user'
	.isheader=1
	.name=Cookie
	
fields.srcdoc.link=codedoc
	.inputfieldname=url
	.fieldrole='doc '
	.isheader=0
	.name=Document

fields.srcdate.link=sortdate
	.inputfieldname=time
	.fieldrole='time'
	.isheader=0
	.name=Timestamp

fields.srcuseragent.link=codeagent
	.inputfieldname=useragent
	.isheader=0
	.name=UserAgent

#this is a dummy field, it will get its input from the sorter
fields.srccount.link=sortcount
	.inputfieldname=time
	.fieldrole='cnt '
	.isheader=1
	.name=QueryCount


#parameters for the encoders
fields.codeuser.link=sortuser
	.precode=1
	.dumptxt=users.txt

fields.codedoc.link=sortdoc
	.precode=1
	.dumptxt=documents.txt

fields.codeagent.link=sortagent
	.precode=0
	.dumptxt=useragents.txt

#parameters for the sorterFields
fields.sortuser.link=outuser
	.sorter=sorter
	.slot=0

fields.sortcount.link=outcount
	.sorter=sorter
	.slot=1

fields.sortdoc.link=outdoc
	.sorter=sorter
	.slot=102

fields.sortdate.link=outdate
	.sorter=sorter
	.slot=100
	.slot=101

fields.sortagent.link=outagent
	.sorter=sorter
	.slot=103

# as csvinput
input.create.lineparser=csvparser
input.lineparser=lineparser
input.lineparser.separator=|
input.create.fieldcollection=fieldcollection
input.fields=fieldcollection
input.fields.input=lineparser

input.fieldcollection.create.ipparser=stringparser
#input.fieldcollection.create.ipparser=ipparser
input.fieldcollection.ipparser.fields=fieldcollection
.source=0
.destination=ipaddress

input.fieldcollection.create.authuser=stringparser
input.fieldcollection.authuser.fields=fieldcollection
.source=1
.destination=authuser

input.fieldcollection.create.date=dateparser
input.fieldcollection.date.fields=fieldcollection
.source=2
.destination=time

input.fieldcollection.create.method=stringparser
input.fieldcollection.method.fields=fieldcollection
.source=3
.destination=method

input.fieldcollection.create.url=stringparser
input.fieldcollection.url.fields=fieldcollection
.source=4
.destination=url

input.fieldcollection.create.param=stringparser
input.fieldcollection.param.fields=fieldcollection
.source=5
.destination=urlparam

input.fieldcollection.create.protocol=stringparser
input.fieldcollection.protocol.fields=fieldcollection
.source=6
.destination=protocol

input.fieldcollection.create.statuscode=stringparser
input.fieldcollection.statuscode.fields=fieldcollection
.source=7
.destination=statuscode

input.fieldcollection.create.length=stringparser
input.fieldcollection.length.fields=fieldcollection
.source=8
.destination=length

input.fieldcollection.create.referer=stringparser
input.fieldcollection.referer.fields=fieldcollection
.source=9
.destination=referer

input.fieldcollection.create.useragent=stringparser
input.fieldcollection.useragent.fields=fieldcollection
.source=10
.destination=useragent

input.fieldcollection.create.origocookie=stringparser
input.fieldcollection.origocookie.fields=fieldcollection
.source=11
.destination=origocookie

input.fieldcollection.create.axelerocookie=stringparser
input.fieldcollection.axelerocookie.fields=fieldcollection
.source=12
.destination=cookie

# input description done.
\endverbatim
    
   */
  void appendFromFile(istream & f);

  //! returns a subkey
  /** Returns NULL if subkey is not found. 
   */
  configuration* getKey(string name);

  //! \brief Displays warning(s) if there are unused configuration options.

  //! It works recursively.

  //! May be used to debug-enumerate configuration if called
  //! right after parsing the config file.
  void warnIfUnused();

  //! Returns a list of all available subkeys.
  vector<pair<string,configuration*> > getSubKeys();
  
 private:
  //! \brief holds the path
  string path;

  //! \brief holds the value(s) associated with this node
  vector<string> values;
  //! \brief hint about unused parameters: stores the index of the last value retrieved
  int lastused;

  //! \brief Are we in an enumeration?
  bool inEnum;
  //! \brief position of enumeration
  int enumPos;

  //! the type of subkey array
  typedef vector<pair<string,configuration*> > subkeys_t;

  //! \brief the subkeys available.
  subkeys_t subkeys;

  //! \brief removes leading and trailing whitespaces from string \p s
  string trim(const string & s);

  //! \brief defines what to be treated as whitespace for \c trim()
  static const char *whitespace;
};


#endif //__CONFIGURATOR_H__
