/******************************************************************************
** Az albbi program s informcik a Nemzeti Kutatsi Fejlesztsi Programok
** 2002. vi plyzatn a 2. sz. programban elfogadott NKFP-2/0017/2002 szm
** ,,Adatrosta - webtrhzak nagytmeg adatbzisainak elemzse adatbnyszati 
** s statisztikai eszkzk segtsgvel'' elnevezs projekt termke.
**
** A program a projekt folyamn feldolgozsra kerl naplllomnyok
** (web, access s mail) binris kdolst s tmrtst vgzi.
**
** A program s informcik az Adatrosta projekt konzorciumnak tulajdona.
** Engedly nlkli felhasznlsa tilos.
**
** Az Adatrosta projekt konzorcium tagjai:
** ELTE TTK Szmtgptudomnyi Tsz.
** BME Matematika Intzet
** MTA SZTAKI Informatikai Kutatlaboratrium
** Axelero Rt.
** econet.hu Rt.
**
** http://adatrosta.hu/
** email: info@adatrosta.hu
**
** (C) 2003. Minden jog fenntartva.
**
**
** This data and information are property of the 'Data Riddle' consortium.
** Modification or use without permission is strictly prohibited.
**
** (C) 2003 All rights reserved.
**
** Author: Balazs RACZ, 2004.
**
****************************************************************************/

/*!
  \file   nonordfp.cpp
  \brief  A frequent itemset miner algorithm using 
          never reordering fp growth algorithm on a compacted FP tree
  \author Balazs RACZ  bracz+ff6@math.bme.hu
  \date   2004-08-04
*/

//#define CHECK


#ifndef NOOPTOUT
#define SINGLET
#endif

//#define FIMI

#define DEFAULT_PROJLEVEL 0
#define DEFAULT_PROJPERCENT 80
#define DEFAULT_DENSESIZE 5000
#define DEFAULT_DENSEPERCENT 70
#define DEFAULT_SIMULTPROJTHRES 1000000

#define CFG

#define FIMIB
//#define CPROFILE
//#define STATS

#define _XOPEN_SOURCE 600

#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <fstream>


#ifdef FIMI
#include "configurator.h"
#if DEBUG_LEVEL>0
#include "log.h"
#else
#define err(x...)
#define info(x...)
#define warn(x...)
#define status(x...)
#endif 
#else
#include "freq/frequisetalgo.h"
#include "lib/log.h"
#endif

#ifdef DISPTIME
#include "lib/progressbar.h"
#else
inline void dispTotalTime() {}
#endif

#include <ext/hash_map>
#include <map>
#include <deque>
#include <vector>
#include <list>
using namespace std;
using namespace __gnu_cxx;

#undef TYPENAME
#define TYPENAME typename

#define FPTREE_RESERVE_SIZE 300000 //approx 75 MB to avoid excessive copies due to vector realloc

#ifdef FIMIB
#define FILLZERO(p,len) memset(p,0,sizeof(p[0])*(len))
#else
#define FILLZERO(p,len) mymemzero(p,len)
#endif

#ifdef CPROFILE
inline long long tsc_read() {
  long long ret;
  asm volatile ("rdtsc" : "=A" (ret));
  return ret;
}
#endif


#define DINLINE __attribute__((always_inline))

/*void __attribute__((noinline)) mymemzero(int *p,size_t len) {
  register int *a=p;
  //len>>=2;
  while (len) {
    len--;
    *a=0;
    a++;
    //    *(a++)=0;
  }
  }*/



#ifndef FIMIB
//typedef int _v2q_t __MODE__((v2q));
typedef int _v2q_t __attribute__ ((mode (V2DI)));

uint64_t totmemzero=0;

void DINLINE mymemzero(int *p,size_t len) {
  register int l=((len<<2)+15)>>4;
  register _v2q_t a = {0,0};
  register _v2q_t *b=(_v2q_t*)p;
  while (l) {
    *b=a;
    b++;
    l--;
  }
  //  memset(p,0,len<<2);
  totmemzero+=len<<2;
}

void __attribute__((always_inline))  mymemzero(uint32_t *p,size_t len) {
  mymemzero((int*)p,len);
}
#endif


#ifdef CPROFILE
//------------ PROFILING ----------------
#define MAXLEVEL 10000

typedef struct levprof_t {
  int currlevel;
  volatile long long lastc;
  long long spent[MAXLEVEL];
  int counter[MAXLEVEL];
  levprof_t() {
    //    FILLZERO((int*)&(spent[0]),MAXLEVEL);	
    memset(spent,0,sizeof(long long)*MAXLEVEL);
    memset(counter,0,sizeof(int)*MAXLEVEL);
    currlevel=0;
    lastc=tsc_read();
  }
} levprof_t;



//#define SETLEVEL(str, level) { long long c=tsc_read(); /*if (c<str.lastc) {warn(0,"tsc goes back: old: %lld, new:%lld",str.lastc,c)} */ str.spent[str.currlevel]+=c-str.lastc; str.lastc=c; str.currlevel=level; }

#define SETLEVEL(x,y) 

#define SETLEVEL_WC(str, level) SETLEVEL(str, level); str.counter[level]++;

#define PROFRESOL (2800000000u)

static levprof_t profrecdepth;

static levprof_t profaggr;
static levprof_t profproj;

#else 

#define SETLEVEL(x,y) 

#define SETLEVEL_WC(str, level)

#endif

//! Memory block allocator class
template<int N, class T> class memoryblocks {
public:
  //! N memory blocks
  typedef struct memblock_t {
    T *mem[N];
  } memblock_t;

  //! request a memory block: N arrays,each of size size length
  memblock_t DINLINE getnextblock(size_t length) {
    if ((nextfree==blocks.end())||(length>nextfree->length)) {
      //allocate block
#ifdef MEMUDEBUG
      if (length>1000000000) {
        err(0,"Length is strange1: %u = %d",length,length);
        exit(2);
      }
#endif      
      smemblock_t newblock;
      length+=length>>1;//allocate somewhat more memory
      for (int j=0;j<N;j++) {
#ifndef FIMIB
	posix_memalign((void **)(newblock.data.mem+j),16,sizeof(T)*length+16);
#else
	newblock.data.mem[j]=(T*)malloc(sizeof(T)*length);
        if (!newblock.data.mem[j]) {
          err(0,"Memory allocation failed! sizeof(T)=%d, length=%d",sizeof(T),length);
          exit(2);
        }
#endif
      }
      newblock.length=length;
      blocks.insert(nextfree,newblock);
      return newblock.data;
    }
    return (nextfree++)->data;
  }
  
  //! release the last non-released block
  void DINLINE freelastblock() {
#ifdef CHECK
    if (nextfree==blocks.begin()) {
      //warn(0,"memoryblocks: freed a non-allocated block!");
      return;
    }
#endif
    nextfree--;
  }

  memoryblocks() {
    nextfree=blocks.end();
  }

  ~memoryblocks() {
    for (TYPENAME blocklist_t::iterator i=blocks.begin();i!=blocks.end();i++) {
      for (int j=0;j<N;j++) {
	free(i->data.mem[j]);
      }
    }
    blocks.clear();
    nextfree=blocks.end();
  }

private:
  //! the data blocks with length
  typedef struct smemblock_t {
    memblock_t data;
    size_t length;
  } smemblock_t;

  //! the list of memory blocks
  typedef list<smemblock_t> blocklist_t;

  //! the actual blocks
  blocklist_t blocks;
  
  //! the next free block	
  TYPENAME blocklist_t::iterator nextfree;
  
};


//! Memory block allocator class
/**
   every new memory block allocated is zeroed, 
   freed blocks have to be zeroed by the caller
*/
template<class T> class zeroingmemoryblock {
public:
  //! request a memory block: N arrays,each of size size length
  T* DINLINE getnextblock(size_t length) {
    if ((nextfree==blocks.end())||(length>nextfree->length)) {
      //allocate block
      smemblock_t newblock;
#ifdef MEMUDEBUG
      if (length>1000000000) {
        err(0,"Length is strange: %u = %d",length,length);
        int a=0;
        err(0,"%d",a);
        a++;
        err(0,"%d",a);
        int *b=NULL;
        *b=3;

      }
#endif      
      length+=length>>1;//allocate somewhat more memory
#ifndef FIMIB 
      posix_memalign((void **)(&newblock.data),16,sizeof(T)*length+32);
#else
      newblock.data=(T*)malloc(sizeof(T)*length);
      if (!newblock.data) {
        err(0,"Memoruy allocation failed! sizeof(T)=%d, length=%d",sizeof(T),length);
        exit(2);
      }
#endif
      FILLZERO(newblock.data,length);
      newblock.length=length;
      blocks.insert(nextfree,newblock);
      return newblock.data;
    }
    return (nextfree++)->data;
  }
  
  //! release the last non-released block
  void DINLINE freelastblock() {
#ifdef CHECK
    if (nextfree==blocks.begin()) {
      //warn(0,"zeroingmemoryblock: freed a non-allocated block!");
      return;
    }
#endif
    --nextfree;
  }

  zeroingmemoryblock() {
    nextfree=blocks.end();
  }

  ~zeroingmemoryblock() {
    for (TYPENAME blocklist_t::iterator i=blocks.begin();i!=blocks.end();i++) {
      free(i->data);
    }
    blocks.clear();
    nextfree=blocks.end();
  }

private:
  //! the data blocks with length
  typedef struct smemblock_t {
    T* data;
    size_t length;
  } smemblock_t;

  //! the list of memory blocks
  typedef list<smemblock_t> blocklist_t;

  //! the actual blocks
  blocklist_t blocks;
  
  //! the next free block	
  TYPENAME blocklist_t::iterator nextfree;
  
};



//! memory allocator class
template<class T, int bigblklen> class singlealloc {
public:
  singlealloc() {
    cbigblk=NULL;
    cfree=0;
  }

  T* DINLINE allocate(int ns) {
    if ((int)cfree<ns) {
      cfree=bigblklen;
      cbigblk=new T[bigblklen];
      bigblocks.push_back(cbigblk);
    }
    T* ret=cbigblk;
    cfree-=ns;
    cbigblk+=ns;
    return ret;
  }

  //! free all allocated blocks
  void freeall() {
    for (TYPENAME bigblocks_t::iterator i=bigblocks.begin();i!=bigblocks.end();i++) {
      delete [] *i;
    }
  }

private:
  typedef list<T*> bigblocks_t; 
  //! the big blocks
  bigblocks_t bigblocks;
  
  //! the current partial big block
  T* cbigblk;
  
  //! the remaining length of the current partial big block
  size_t cfree;
};

//! memory array allocator class
template<class T,int base, int bigblklen> class arrayalloc{
public:
  static  size_t DINLINE BLOCKLEN(int size)  {
    return (base << size);
  }

  //! allocate a BLOCKLEN(size) sized block
  T* DINLINE allocate(unsigned int size) {
    if (freelists.size()<=size) {
      freelists.resize(size+1);
    }
    T* ret;
    if (ret=freelists[size]) {
      //we have a cached block
      T** d=(T**)ret;
      freelists[size]=*d;
      return ret;
    } else {
      //allocate a new block      
      size_t ns=BLOCKLEN(size);
      return thealloc.allocate(ns);
    }
  }//allocate

  //! free a BLOCKLEN(size) sized block
  void DINLINE free(T* data,unsigned int size) {
    if (freelists.size()<=size) {
      freelists.resize(size+1);
    }
    T** d=(T**)data;
    *d=freelists[size];
    freelists[size]=data;
  }//free
  
  
  arrayalloc() :thealloc(){
  }

private:

  //! the list of freed small blocks
  vector<T*> freelists;

  //! the actual allocator
  singlealloc<T,bigblklen> thealloc;

};




template<class key, class value> class linmap {
public:
  typedef key key_type;
  typedef value data_type;
  typedef pair<key_type, data_type> value_type;

  typedef std::allocator<linmap<key_type,value_type> > allocator;

  linmap() {
    data=NULL;
    size=0;
    logsize=-1;
  }

  data_type & DINLINE  operator[](key_type k) {
    unsigned int pos=0;
    while((pos<size)&&(data[pos].first!=k)) {
      pos++;
    }
    if (pos>=size) {
      //insert
      size_t dlen=size?myallocator.BLOCKLEN(logsize):0;
      if (pos>=dlen) {
	//realloc
	value_type *nd=myallocator.allocate(logsize+1);
	if (size) {
	  sort(data,data+size);
	  memcpy(nd,data,size*sizeof(value_type));
	  myallocator.free(data,logsize);
	}
	logsize++;
	data=nd;
      }
      data[pos].first=k;
      data[pos].second=data_type();
      size++;
    }
    //swap with the previous
    /*if (pos) {
      swap(data[pos-1],data[pos]);
      pos--;
      }*/
    return data[pos].second;
  }

  class iterator {
  private:
    size_t pos;
    linmap<key_type,data_type> *parent;
  public:
    value_type& DINLINE operator*()  {
      return parent->data[pos];
    }
    
    value_type* DINLINE operator->()  {
      return parent->data+pos;
    }

    void DINLINE operator++ ()  {
      pos++;
      //return *this;
    }
    //! bug: iterators of different base may be equal
    bool DINLINE  operator==(const iterator &o) {
      return pos==o.pos;
    }

    bool DINLINE operator!=(const iterator &o)  {
      return pos!=o.pos;
    }
    
    inline iterator(linmap<key_type,data_type> *p,size_t s){
      parent=p;
      pos=s;
    }
  };
  friend class iterator;

  iterator DINLINE begin()  {
    return iterator(this,0);
  }

  iterator DINLINE end()  {
    return iterator(this,size);
  }

private:
  //! the actual length;
  uint32_t size;
  
  //! blocksize of the data space
  int logsize;

  //! the actual data space
  value_type *data;

  //! allocator for the blocks
  typedef arrayalloc<value_type,5,100000> myallocator_t;

  static myallocator_t myallocator;

};

template<class key, class value> TYPENAME linmap<key,value>::myallocator_t linmap<key,value>::myallocator;

//! NonOrdFP miner algo
/**
   Accepted configuration options: 
   
   \param simultprojthres if the fp tree's size is above this
   threshold, then use simultaneous projection on the first level.

   \param projpercent if the number of deleted FP tree nodes exceeds
   this percentage, then projection is calculated.

   \param densepercent if the number of deleted FP tree nodes exceeds
   this percentage, then sparse aggregation is used.

   \param projlevel all the projections until this level are done. 

   \param filterdupl if non-zero, filter duplicate items from baskets

   (projlevel 0 and projpercent 101 means do not do any projections)
*/
#ifdef FIMI

class NonOrdFp;

//! type of item identifier
typedef int item_t;

//! itemset (in a vector)
typedef vector< item_t > item_vector_t;


#define DEFAULT_BUFSIZE (3*1024*1024)

#define LINEBUFLEN 5000

//#define DBGG

//! A coordinator for FIMI frequent itemset mining
/**
   NOTE: us eonly algorithms that do not need the pipeline (as there
   is no pipeline in this case).

   Accepted configuration options:

   \param filename the output file name, if omitted, no output is written.

   \param algo an object name, of type FrequisetAlgo to do the actual calculations

   \param sourcefilename the input file name, required. FIMI format: each transaction is one line, separated by non-digits, \n ends line, last line needs an ending \n, too.

   \param bufsize is an integer, the amount of buffer (in bytes) to allocate for stdio FILE input/output. Default: DEFAULT_BUFSIZE =3 MB.

*/
class FIMIcoord {
public:

  void applyConfig(string filename, string infilename);

  void doCoding();

  //! callback function from the FrequisetAlgo
  void DINLINE pushFreqItems(size_t support, item_vector_t &items);

  //! stack based output: push onto the stack and output the current
  void DINLINE pushItem(size_t support,item_t item);

  //! pop stack
  void DINLINE popStack();

  //! display histogram of frequent itemsets
  void dispHisto() {
    info(0,"Histogram of frequent itemset length:");
    int i=1;
    while ((i<maxitem)&&(histo[i])) {
      printf("%d\n",histo[i++]);
    }
  }


public: //originally was private:
  
  //! The input file name
  string infilename;
  
  //! The output file name
  string filename;

  //! The input file descriptor
  FILE *fin;

  //! The output file descriptor
  FILE *fout;
  int fdout;

  //! The implementation of FrequISet algorithm
  NonOrdFp * algo;

  //! buffer for output file
  char *outbuf, *buffree, *bufend;

  //! buffer for input file
  char *inbuf;

  //! buffer size;
  size_t bufsize;

  void pushAllRows();

  //! the number of frequent items found
  long long numfreq;

  //! string version of the frequent items
  typedef hash_map<item_t,string> itemsmap_t;

  itemsmap_t fitems;

  //! the stack of frequent itemsets
  //stack<int> endposst;
  int *endposst;
  size_t endpossp;

  //! histogram of frequent itemsets
  int *histo;

  //! the largest item encountered
  item_t maxitem;

  //! text representation of items
  char *itemtt;

  //! the buffer in which to render output lines
  char *linebuf;
  
  //! buffered file write
  void bufdwrite(char *buf,size_t len);

  //! buffered file flush
  void bufdflush();

  void putend(char *buf,item_t item,int &nextt,size_t support,char **last);
};





void DINLINE FIMIcoord::pushItem(size_t support,item_t item) {
  /*
  string &r=fitems[item];
  if (r.size()) {
    //we have text version
  } else {//we don't have text version
    
  }
  */
  if (fdout) {
    int nextt=0;
    //sprintf(linebuf+endposst.top(),"%d %n (%d)\n",item,&nextt,support);
    char *n;
    size_t lp=endposst[endpossp];
    putend(linebuf+lp,item,nextt,support,&n);
    endpossp++;
    endposst[endpossp]=lp+nextt;
    //    endposst.push(endposst.top()+nextt);
    //fputs(linebuf,fout);
    /*int pos=0;
    int len=n-linebuf;
    while (pos<len) {
      int ret= write(fdout,linebuf+pos,len-pos);
      if (ret<0) {
	perror("write");
      } else 
	pos+=ret;
	}*/
    //    fwrite(linebuf,n-linebuf,1,fout);
    bufdwrite(linebuf,n-linebuf);
  } else {
    endpossp++;
  }
  histo[endpossp]++;
  numfreq++;
}


void DINLINE FIMIcoord::popStack() {
  //  if (fdout)
    //endposst.pop();
    endpossp--;
}




class NonOrdFp {
public:
  size_t min_supp;
  FIMIcoord *output;
#else
class NonOrdFp: public DefFrequisetAlgo {
#endif
private:
  static const item_t npos=(item_t)-1;

  // BASIC TYPES
  //! index type of the second trie
  typedef uint32_t nodeindex_t;

  //! counter
  typedef uint32_t counter_t;


  //------------- SECOND TREE ------------

  //! the parents
  //typedef vector<nodeindex_t> parents_t;
  typedef nodeindex_t* parents_t;

  //! the counters
  //typedef vector<counter_t> counters_t;
  typedef counter_t* counters_t;

  //! item offsets
  //typedef vector<nodeindex_t> itemstarts_t;
  typedef nodeindex_t* itemstarts_t;
  
  //! item counters
  //typedef vector<counter_t> itemcounters_t;
  typedef counter_t* itemcounters_t;

  // the main containers of the second tree
  parents_t mtparents;
  counters_t mtcounters;
  itemstarts_t mtitemstarts,tempitemstarts;
  itemcounters_t mtitemcounters;

  //!allocator for itemcounters
  memoryblocks<1,nodeindex_t> allocitemstarts;

  //! allocator for counters
  zeroingmemoryblock<counter_t> alloccounters;

  //! allocator for projections
  typedef memoryblocks<2,nodeindex_t> projallocator_t;
  projallocator_t allocprojections;

  //---------- PERSISTENT MEMORY -------
  
  //! npos
  parents_t nposit;

  //sparse aggregation memory
  parents_t nonzerolists;
  itemstarts_t heads,tails;


  //------------- MISC -----------------

  //! type of passes we can have
  enum pass_t {
    init,  //no pass whatsoever
    build_histogram, //fill the histogram data
    build_fptree     //build the FP tree
  };

  //!which pass are we in?
  pass_t cpass;


  //! our histogram collect datatype
  //typedef hash_map<item_t, uint32_t> hist_t;
  typedef vector<uint32_t> hist_t;

  //! histogram built by us
  hist_t myhist;

  //! the total count of transactions
  uint32_t transcount;

  //! type of order map
  //typedef hash_map<item_t,item_t> order_map_t;
  typedef vector<item_t> order_map_t;

  //! reorder map
  order_map_t order;
  
  //! inverse of reorder map
  vector<item_t> backorder;


  //! item count
  uint32_t itemcount;

  //------------- BUILD FP TREE -----------------
  //! branch of the fp tree
  typedef linmap<item_t, uint32_t> tree_child_t;
  
  //! the children of all the nodes of the fp tree
  //typedef deque<tree_child_t> trie_children_t;

  //! an element of the fp tree
  typedef struct fp_item_t {
    //    item_t *begin,*end;
    item_t item;
    uint32_t count;
    uint32_t parent;
    tree_child_t child;
    //    fp_item_t(item_t*b, item_t *e,uint32_t p) {
    fp_item_t(item_t i,uint32_t p) {
      //begin=b;end=e;
      item=i;
      parent=p;
      count=0;
    }
  } fp_item_t;

  //! the entire fp tree
  typedef vector<fp_item_t> fptree_build_t;
    

  //! the entire fp tree
  typedef vector<fp_item_t> fptree_t;

  //! basket trie downstream
  //trie_children_t fptree_d;
  
  //! basket trie upstream (for recursive copy)
  fptree_t fptree;

  //! basket trie upstream (build)
  fptree_build_t fptreeb;

  //! temporary memory to store the database
  singlealloc<item_t,10000> dballoc;

  //! database transactions
  typedef deque<item_t *> fulldb_t;
  fulldb_t fulldb;

  //!projection config: percent of nulls after which do the projection
  int projpercent;

  //!projection config: level until do all the projections
  int projlevel;

  //! projection config: fp tree size (in nodes) above which to do simult projection
  unsigned int simultprojthres;

  //! config: percent of nulls below which to switch to dense aggregation
  int densepercent;

  //! config: fp tree size below which to switch to dense aggregation
  int densesize;

  //! config: filter duplicates?
  bool filterdupl;

  //! config: do sparse aggregates?
  bool sparseaggr;
#ifdef STATS
  //! statistics: number of projections done
  size_t numproj;
  //! statistics: number of recursive calls (sparse)
  size_t numsp;
  //! statistics: number of recursive calls (dense)
  size_t numden;

  //! statistics: sparse->dense switch
  size_t sp2den;
  //! statistics: sparse->project 
  size_t sp2proj;
  //! statistics: dense->sparse 
  size_t den2sp;
  //! statistics: dense->proj 
  size_t den2proj;
  //! statistics: dense->singlet 
  size_t den2sing;
  //! statistics: sparse->singlet 
  size_t sp2sing;

  //! statistics: number of total fp tree nodes after projections
  int64_t projtotlen;

  //! statistics: number of total fp tree nodes in projections
  int64_t projinlen;

  //! statistics: total fptree node count on which aggregate was called
  int64_t aggrtotlen;
  
  //! statistics: number of occasions where single-chain optimization could be used
  int singltcount;

  //! statistics: total node count where single-chaing optimization could be used
  int singltnodes;
#endif

public:
  
  virtual ~NonOrdFp() {}

  NonOrdFp(){cpass=init;};
  
  void custApplyConfig(configuration *config) {
    configuration *key;    
    key=config->getKey("projpercent");
    if ((!key)||(key->startEnum(),key->endEnum())) {
      projpercent=DEFAULT_PROJPERCENT;
    } else {
      projpercent=key->getValue_range(0,101);
      info(0,"Projecting above %d percent empty cells.",projpercent);
    }

    key=config->getKey("densepercent");
    if ((!key)||(key->startEnum(),key->endEnum())) {
      densepercent=DEFAULT_DENSEPERCENT;
    } else {
      densepercent=key->getValue_range(0,101);
      info(0,"Switching to dense below %d percent empty cells.",densepercent);
    }

    key=config->getKey("densesize");
    if ((!key)||(key->startEnum(),key->endEnum())) {
      densesize=DEFAULT_DENSESIZE;
    } else {
      densesize=key->getValue_int();
      info(0,"Switching to dense below %d cells.",densesize);
    }
    
    key=config->getKey("projlevel");
    if ((!key)||(key->startEnum(),key->endEnum())) {
      projlevel=DEFAULT_PROJLEVEL;
    } else {
      projlevel=key->getValue_int();
      info(0,"Projecting until level %d",projlevel);
    }

    key=config->getKey("simultprojthres");
    if ((!key)||(key->startEnum(),key->endEnum())) {
      simultprojthres=DEFAULT_SIMULTPROJTHRES;
    } else {
      simultprojthres=key->getValue_int();
      info(0,"Using simultaneous projection above fp tree size %d",simultprojthres);
    }

    key=config->getKey("filterdupl");
    if ((!key)||(key->startEnum(),key->endEnum())) {
      filterdupl=false;
    } else {
      filterdupl=(key->getValue_int()!=0);
      if (filterdupl) {
	info(0,"Filtering duplicates");
      } else {
	info(0,"Not filtering duplicates");
      }
    }
    
    key=config->getKey("sparseaggr");
    if ((!key)||(key->startEnum(),key->endEnum())) {
      sparseaggr=false;
    } else {
      sparseaggr=(key->getValue_int()!=0);
      if (sparseaggr) {
	info(0,"Using sparse aggregation");
      } else {
	info(0,"Not using sparse aggregation");
      }
    }


  }
  
  void startPass() {
    switch(cpass) {
    case init: {
      cpass=build_histogram;      
    }
    case build_histogram: 
      info(0,"Building histogram.");
      transcount=0;
      break;
    case build_fptree:
      info(0,"Building fp tree.");
      //initialize root node of fp tree
      fptreeb.reserve(FPTREE_RESERVE_SIZE);
      fptreeb.push_back(fp_item_t(npos,0));
      //      fptree_d.push_back(tree_child_t());
      break;
    }
  }

  void treebPass() {
    for (fulldb_t::iterator ct=fulldb.begin();ct!=fulldb.end();++ct) {
      item_t *items=*ct+1;
      unsigned int ii=0;
      for (item_t i=0;i<**ct;i++) {
	//order_map_t::iterator j=order.find(items[i]);
	register item_t ci=items[i];
	if (myhist[ci]>=min_supp) {
	  items[ii++]=order[ci];
	}
	/*	if (j!=order.end()) {
	  items[ii++]=j->second;
	  }*/
      }
      **ct=ii;
      //items.erase(items.begin()+ii,items.end());
      //reverse sort for freqency-ascending order
      //sort(items.begin(),items.end(),greater<item_t>());
      sort(items,items+ii);
      if (filterdupl) {
	item_t *l=unique(items,items+ii);
	ii=**ct=l-items;
	//items.erase(l,items.end());
      }
      //      items.uniq();
      //insert into fp tree
      uint32_t fpidx=0;
      fptreeb[fpidx].count++;//increase counter	
      for (item_t i=0;i<**ct;i++) {
	uint32_t &nn=fptreeb[fpidx].child[items[i]];
	if (!nn) { //insert node
	  uint32_t nidx=fptreeb.size();
	  fptreeb.push_back(fp_item_t(items[i],fpidx));
	  //fptree_d.push_back(tree_child_t());
	  nn=nidx;
	}
	fpidx=nn;
	fptreeb[fpidx].count++;//increase counter
      }
    }
  }


  void DINLINE In(item_vector_t &items) {
    switch(cpass) {
      
    case build_histogram: {
      transcount++;
      //filter dupl
      if (filterdupl) {
	sort(items.begin(),items.end());
	item_vector_t::iterator l=unique(items.begin(),items.end());
	items.erase(l,items.end());
      }
      for (unsigned int i=0;i<items.size();i++) {
	if (items[i]>=(int)myhist.size()) {
	  myhist.resize((items[i]+1)<<1);
	}
	myhist[items[i]]++;
      }
      item_t *b=dballoc.allocate(items.size()+1);
      *b=items.size();
      memcpy(b+1,&(items[0]),items.size()<<2);
      fulldb.push_back(b);
      return;
      break;
    }//case histogram
    case build_fptree: {
      break;
    }
    default:
      err(0,"In() in an unknown pass %d!",cpass);
      break;
    }
  } //proc In
  
  bool finishPass() {
    switch(cpass) {
    case build_histogram: {
      calcreorder();
      cpass=build_fptree;
      dispTotalTime();
      startPass();
      treebPass();
      dballoc.freeall();
    }
    case build_fptree: {
      item_vector_t iv;
      info(0,"Computing frequent itemsets.");
#ifdef STATS
      numsp=numden=sp2den=den2sp=sp2proj=den2proj=den2sing=sp2sing=0;
      projtotlen=0;
      projinlen=0;
      aggrtotlen=0;
      singltcount=0;
      singltnodes=0;
#endif

      //FILLZERO(nonzerolists,nexto);
      heads=allocitemstarts.getnextblock(itemcount+1).mem[0];    
      tails=allocitemstarts.getnextblock(itemcount+1).mem[0];    

      info(0,"transcount %d",transcount);
      output->pushFreqItems(transcount,iv);
      if (fptreeb.size()>simultprojthres) {
	copyfptree();
	simultproject(0,itemcount,mtparents,mtcounters,
		      mtitemstarts,mtitemcounters);
      } else {
	copyfptree();
	dononprojfp_sparse(0,itemcount,mtparents,mtcounters,
			   mtitemstarts,mtitemcounters,transcount,1);
      }
#ifdef STATS
      info(0,"Done. %d sp, %d den calls, %d sp %d den projections (len %lld), total new node count %lld, total aggregate %lld, total zerofill %lld, singlt: count=%d sp %d den nodes=%d",numsp,numden,sp2proj,den2proj, projinlen, projtotlen, aggrtotlen,totmemzero,sp2sing,den2sing,singltnodes);
      info(0,"sp2den %d, sp2proj %d, sp2sing %d; den2sp %d, den2proj %d, den2sing %d",sp2den,sp2proj,sp2sing,den2sp,den2proj,den2sing);
#else
      info(0,"Done.");
#endif
      return false;
      break;
    }
    case init: {
      warn(0,"finishPass in init! strange!");
    }
    }
    return false;
  }

  //unused currently
  void finish() {
#ifdef CPROFILE
    displayprof(profrecdepth,"Recursion level");
    displayprof(profaggr,"cpb-aggregate");
    displayprof(profproj,"project");
#endif
  }
  

private:

#ifdef CPROFILE
  void displayprof(levprof_t &p,string d) {
    fprintf(stderr,"Profile: %s (resolution: %u) (measures REAL time)\n",d.c_str(),PROFRESOL);
    int maxl=0;
    for (int i=0;i<MAXLEVEL;i++) {
      if (p.spent[i])
	maxl=i;
    }
    long long total=0;
    for (int i=0;i<=maxl;i++) {
      fprintf(stderr,"Level %3d: %.6lf seconds %u count\n",i,double(p.spent[i])/PROFRESOL,p.counter[i]);
      total+=p.spent[i];
    }
    fprintf(stderr,"Total: %.6lf seconds\n",double(total)/PROFRESOL);
    fprintf(stderr,"Non-0: %.6lf seconds\n",double(total-p.spent[0])/PROFRESOL);
  }
#endif

  void reccopytree(nodeindex_t cnode,nodeindex_t parent) {
    nodeindex_t pos=0;
    fp_item_t *ci=&(fptree[cnode]);
    if (cnode) 
      pos=tempitemstarts[ci->item]++;
    mtparents[pos]=parent;
    mtcounters[pos]=ci->count;
    for(tree_child_t::iterator i=ci->child.begin();i!=ci->child.end();++i) {
      reccopytree(i->second,pos);
    }
  }

  void copyfptree() {
    dispTotalTime();
    info(0,"Copying fp tree. (Node count=%u)",fptreeb.size());
    info(0,"Phase 1");
    //fptree.reserve(fptreeb.size());
    //fptree.insert(fptree.end(),fptreeb.begin(),fptreeb.end());
    fptree.swap(fptreeb);
    info(0,"Clearing build fp tree.");
    fptreeb.clear();
    info(0,"Creating second tree.");
    mtitemstarts=allocitemstarts.getnextblock(itemcount+1).mem[0];
    mtitemcounters=allocitemstarts.getnextblock(itemcount+1).mem[0];
    tempitemstarts=allocitemstarts.getnextblock(itemcount+1).mem[0];
    FILLZERO(mtitemstarts,itemcount+1); 
    FILLZERO(tempitemstarts,itemcount+1);
    FILLZERO(mtitemcounters,itemcount+1);
    //mtitemstarts.resize(itemcount+1);
    //mtitemcounters.resize(itemcount);
    //first pass: calculate item counts
    for (fptree_t::iterator i=++fptree.begin();i!=fptree.end();i++) {
      mtitemstarts[i->item]++;
      mtitemcounters[i->item]+=i->count;
    }
    //aggregate for item start counts
    nodeindex_t tmp,nexto=1;
    for(item_t i=0;i<=(int)itemcount;i++) {
      tmp=nexto;
      nexto+=mtitemstarts[i];
      mtitemstarts[i]=tmp;
      tempitemstarts[i]=tmp;
    }
    //second pass: do the reallocation and the actual copy
    projallocator_t::memblock_t nm=allocprojections.getnextblock(nexto);
    mtparents=nm.mem[0];
    FILLZERO(mtparents,nexto);
    mtcounters=nm.mem[1];
    FILLZERO(mtcounters,nexto);

    //do global allocations
    nposit=alloccounters.getnextblock(nexto);
    //FILLZERO(nposit,nexto);
    nonzerolists=alloccounters.getnextblock(nexto);


    //    mtparents.resize(nexto);
    //mtcounters.resize(nexto);
    //    itemstarts_t ssave(mtitemstarts);
    reccopytree(0,0);
    //mtitemstarts.swap(ssave);
    info(0,"Clearing build fp tree.");
    fptree_t ft;
    fptree.swap(ft);
    ft.clear();
    //fptree_d.clear();
    info(0,"Done.");
  }

  //! do the reordering of the items
  void calcreorder() {
    //    if (reorder==no_change) {
    //  return;
    //}
    info(0,"Reordering items...");
    vector<pair<uint32_t,item_t> > freqs;
    freqs.reserve(myhist.size());
    /*    for(hist_t::iterator i=myhist.begin();i!=myhist.end();i++) {
      if (*i>=min_supp)//filter infrequent items
	freqs.push_back(make_pair(*i->second,i->first));
	}*/
    for (item_t i=0;i<(int)myhist.size();i++) {
      if (myhist[i]>=min_supp) {
	freqs.push_back(make_pair(myhist[i],i));
      }
    }
    //    if (reorder==freq_asc) {
    //sort(freqs.begin(),freqs.end());
    //} else if (reorder==freq_desc) {
    sort(freqs.begin(), freqs.end(), greater< pair<uint32_t,item_t> >() );
      //}
    itemcount=freqs.size();
    backorder.resize(itemcount);
    order.resize(myhist.size());
    for (unsigned i=0;i<freqs.size();i++) {
      order[freqs[i].second]=i;
      backorder[i]=freqs[i].second;
    }
    info(0,"done. %d %d",backorder.size(),order.size());
    info(0,"reordering histogram...");
    /*
    myhist.clear();
    
    for(unsigned i=0;i<freqs.size();i++) {
      if (freqs[i].first>=min_supp) //filter infrequent items
	myhist.insert(make_pair(i,freqs[i].first));
	}*/
    info(0,"done. %d %d",backorder.size(),order.size());    


  }



  //------------ SIMULTANEOUS PROJECTION ----------
  typedef struct projelement_t {
    item_t item;
    uint32_t counter;
    projelement_t *next;
    projelement_t * DINLINE insertchain(item_t _i,uint32_t _c,projelement_t *_n) {
      item=_i;
      counter=_c;
      next=_n;
      return this;
    }
  } projelement_t;
  
  typedef projelement_t * projnode_t;
  typedef vector<projnode_t> projtree_t;

  typedef singlealloc<projelement_t,100000> spalloc_t;

  void __attribute__((noinline)) simultproject(int recdepth,item_t lastitem,parents_t parents,counters_t counters,itemstarts_t itemstarts,itemcounters_t itemcounters) {
    projtree_t tree;
    vector<itemcounters_t> nitemcounters;
    vector<itemstarts_t> nitemstarts;
    vector<counters_t> ncounters;
    vector<parents_t> nparents;
    spalloc_t myalloc;
    tree.resize(itemstarts[lastitem],NULL);
    nitemcounters.resize(lastitem);
    nitemstarts.resize(lastitem);
    ncounters.resize(lastitem);
    nparents.resize(lastitem);
    //PHASE 0: allocate (back-to-forth)
    for (item_t i=lastitem-1;i>=0;i--) {
      itemcounters_t c=allocitemstarts.getnextblock(i+1).mem[0];
      FILLZERO(c,i+1);
      nitemcounters[i]=c;
      itemstarts_t cc=allocitemstarts.getnextblock(i+1).mem[0];
      FILLZERO(cc,i+1);
      nitemstarts[i]=cc;
    }
    info(0,"Simult aggregate...");
    SETLEVEL_WC(profaggr,recdepth+1);

    //PHASE 1: aggregate
    for (item_t i=lastitem-1;i>=0;i--) {
      for (nodeindex_t n=itemstarts[i];n<itemstarts[i+1];n++) {
	//process node n of item i
	//insert current item to current node
	if (itemcounters[i]>=min_supp){
	  projelement_t *ne=myalloc.allocate(1);
	  tree[n]=ne->insertchain(i,counters[n],tree[n]);
	  //	  cn.push_back(make_pair(i,counters[n]));
	}
	projnode_t *pn=&(tree[parents[n]]);
	register projnode_t ccn=tree[n],ppn;
	item_t citem,pitem;
	while ((ccn)&&(*pn)) {
	  ppn=*pn;
	  citem=ccn->item; pitem=ppn->item;
	  if (citem==pitem) {
	    //eq --> sum
	    //npn.push_back(make_pair(citem,cn[ci].second+pn[pi].second));
	    (ppn)->counter+=(ccn)->counter;
	    //condition: cnci.first, item:i, counter=cnci.second
	    nitemcounters[citem][i]+=ccn->counter;
	    //also a node is needed 
	    nitemstarts[citem][i]++;
	    ccn=(ccn->next);//ci++;
	    pn=&(ppn->next);//pi++;
	  } else if (citem>pitem) {
	    //npn.push_back(pn[pi]);
	    pn=&(ppn->next);//pi++;
	  } else {
	    projelement_t *ne=myalloc.allocate(1);
	    *pn=ne->insertchain(ccn->item,ccn->counter,ppn);
	    //npn.push_back(cn[ci]);
	    //condition: cnci.first, item:i, counter=cnci.second
	    nitemcounters[citem][i]+=ccn->counter;
	    //also a node is needed
	    nitemstarts[citem][i]++;
	    ccn=(ccn->next);//ci++;
	    pn=&((*pn)->next);////itt baj lehet a pn-nel!
	  }
	}//merge
	while (ccn) {
	  //ccn=*cn;
	  projelement_t *ne=myalloc.allocate(1);
	  *pn=ne->insertchain(ccn->item,ccn->counter,*pn);
	  //npn.push_back(cn[ci]);
	  item_t citem=ccn->item;
	  //condition: cnci.first, item:i, counter=cnci.second
	  nitemcounters[citem][i]+=ccn->counter;
	  //also a node is needed
	  nitemstarts[citem][i]++;
	  ccn=(ccn->next);//ci++;
	  pn=&((*pn)->next);
	}
	/*while (pi<pn.size()) {
	  npn.push_back(pn[pi++]);
	}
	pn.swap(npn);*/
      }
    }
    info(0,"Simult length...");
    // PHASE 2: set new itemstarts, calculate lengths, allocate new counters
    //set itemstarts, filter infrequent items
    for (item_t i=lastitem-1;i>=0;i--) {
      if (itemcounters[i]<min_supp)
	continue;
      //condition:i
      itemcounters_t icounts=nitemcounters[i];
      nodeindex_t nn=1;//reserve node 0
      itemstarts_t istarts=nitemstarts[i];
      for(item_t ci=0;ci<i;ci++) {
	if (icounts[ci]>=min_supp) {
	  //frequent
	  nodeindex_t cnt=istarts[ci];
	  istarts[ci]=nn;
	  nn=istarts[ci]+cnt;
	} else {
	  //infrequent
	  istarts[ci]=nn;
	}
      }//for current item
      istarts[i]=nn;//set total length
      //allocate counter -- cycle goes back to front
      projallocator_t::memblock_t nm=allocprojections.getnextblock(nn);
      ncounters[i]=nm.mem[0];
      nparents[i]=nm.mem[1];
      nparents[i][0]=0;
      ncounters[i][0]=0;
    }//for item condition

    SETLEVEL_WC(profaggr,0);

    info(0,"Simult project...");
    SETLEVEL_WC(profproj,recdepth+1);
    // PHASE 3: PROJECT
    vector<nodeindex_t> nns;
    nns.resize(lastitem,1);
    //initialize pos of conditional trees
    for (projelement_t *cn=tree[0];cn;cn=cn->next) {
      cn->counter=0;//set parent->pos
    }

    for (item_t i=0;i<lastitem;i++) {
#if 0  //error checking
      for (item_t cond=i;cond<lastitem;cond++) {
	//	if (nitemcounters[cond][i]<min_supp)
	//  continue;
	if (nns[cond]!=nitemstarts[cond][i]) {
	  err(0,"in project: cond=%d, i=%d, nns[cond]=%d, nitemstarts[cond][i]=%d.",cond,i,nns[cond],nitemstarts[cond][i]);
	}
      }
#endif
      for (nodeindex_t n=itemstarts[i];n<itemstarts[i+1];n++) {
	register projnode_t cn=tree[n];
	register projnode_t pn=tree[parents[n]];
	for (;cn;cn=cn->next) {
	  item_t citem=cn->item;
	  if (citem==i)
	    continue;   //this is the condition, no node is needed
	  //now: with condition cn[ci].first we have counter=cn[ci].second
	  while (pn->item<citem)  //the conditions in the parent is always 
	                              //a superset of the conditions in the 
	                              //current node
	    pn=pn->next;
#if 0  //error checking
	  if (pn->item!=citem) {
	    err(0,"Parent is not a superset of the current node!. n=%d, p=%d",n,parents[n]);
	  }
#endif
	  if (nitemcounters[citem][i]>=min_supp) {
	    nparents[citem][nns[citem]]=pn->counter;  //read parent's position
	    ncounters[citem][nns[citem]]=cn->counter; //read counter
	    cn->counter=nns[citem]++;                 //save position
	  } else {
	    //infrequent
	    cn->counter=pn->counter;                //save position
	  }
	}//process the conditions in the current node
      }//for node
      /*//check itemstarts
      for (item_t citem=i+1;citem<lastitem;citem++) {
	if (itemcounters[citem]<min_supp)
	  continue;
	if(nns[citem]!=nitemstarts[citem][i+1]) {
	  warn(0,"Item start mismatch: citem=%d, nns[citem]=%d, i=%d,nitemstarts[citem][i+1]=%d",citem,nns[citem],i,nitemstarts[citem][i+1]);
	}
	}*/
    }//for item

    SETLEVEL_WC(profproj,0);
    myalloc.freeall();

    info(0,"Simult recurse...");
    dispTotalTime();
    // PHASE 4: recurse (front-to-back)
    for (item_t i=lastitem-1;i>=0;i--){//(int)lastitem;i++) {
      //recursive call
      if (itemcounters[i]<min_supp)
	continue;

      //freqset.push_back(backorder[i]);
      SETLEVEL(profrecdepth,0);	
#ifndef NOOPTOUT
      output->pushItem(itemcounters[i],backorder[i]);
#else
      output->pushFreqItems(itemcounters[i],freqset);
#endif
      dononprojfp_sparse(recdepth+1,i,nparents[i],ncounters[i],nitemstarts[i],nitemcounters[i],itemcounters[i],1);
      //freqset.pop_back();
      SETLEVEL(profrecdepth,recdepth+1);
#ifndef NOOPTOUT
      output->popStack();
#endif

      //allocprojections.freelastblock();
      //allocitemstarts.freelastblock();//itemstarts
      //allocitemstarts.freelastblock();//itemcounters
    }
    info(0,"FInish.");
  }

  
  // a local return value of cpb_aggregate: true if each frequent item has at most one node
  //! the last item index which has at most one node is lastsinglet-1.
  int lastsinglet;

  size_t  DINLINE cpb_aggregate_dense(item_t i,
				      itemstarts_t itemstarts,
				      counters_t ncounters,
				      counters_t counters,
				      parents_t parents,
				      itemcounters_t nitemcounters
				      ) {
    size_t nullcount=0;
#ifdef STATS
    aggrtotlen+=itemstarts[i+1];
#endif
    //fill last level of counters
    for(nodeindex_t n=itemstarts[i];n<itemstarts[i+1];n++) {
#ifdef CHECK
      if (parents[n]>=itemstarts[i]) {
	warn(0,"1parent[%d]=%d, > %d",n,parents[n],itemstarts[i]);
      }
#endif  
      if (!counters[n])
	continue;
      register int p=parents[n];
      ncounters[p]+=counters[n];
    }
    //fill upwards counters
#ifdef SINGLET
    lastsinglet=i;
#endif
    for (item_t ci=i-1;ci>=0;ci--) {
      size_t incount=0;
      for(nodeindex_t n=itemstarts[ci];n<itemstarts[ci+1];n++) {
	if (ncounters[n]) {
#ifdef CHECK
	  if (parents[n]>=itemstarts[i]) {
	    warn(0,"2parent[%d]=%d, > %d",n,parents[n],itemstarts[i]);
	  }
	  if (parents[n]>=n) {
	    warn(0,"3parent[%d]=%d, > n: %d",n,parents[n],n);
	  }
	  if (parents[n]>=itemstarts[ci]) {
	    warn(0,"4parent[%d]=%d, > cistart: %d",n,parents[n],itemstarts[i]);
	  }
#endif      
	  ncounters[parents[n]]+=ncounters[n];
	  nitemcounters[ci]+=ncounters[n];
	} else {
	  nullcount++;
	  incount++;
	}//if the counter is non-zero
      }//for items of current level
      int lposcount=itemstarts[ci+1]-itemstarts[ci]-incount;
      if (nitemcounters[ci]<min_supp) {
	nullcount+=lposcount;
      } 
#ifdef SINGLET
      else {
	if (lposcount>1) {
	  lastsinglet=ci;
	}
      }
#endif
    }//for upper levels
    /*if (sparseaggr) {
      alloccounters.freelastblock();
      allocitemstarts.freelastblock();
      }*/
    return nullcount;
  }

  void DINLINE  cpb_clear(item_t i,counters_t ncounters,parents_t parents,itemstarts_t itemstarts) {
    for(nodeindex_t n=itemstarts[i];n<itemstarts[i+1];n++) {
      register nodeindex_t nn=parents[n];
      while ((nn) && (ncounters[nn])) {
	ncounters[nn]=0;
	nn=parents[nn];
      }
    }//for items on the last level
  }



  size_t  DINLINE cpb_aggregate_sparse(item_t i,
				       itemstarts_t itemstarts,
				       counters_t ncounters,
				       counters_t counters,
				       parents_t parents,
				       itemcounters_t nitemcounters
				       ) {
    size_t nullcount=0;
#ifdef STATS
    aggrtotlen+=itemstarts[i+1];
#endif
    FILLZERO(heads,i+1);
    //fill last level of counters
    item_t citem=0;
    for(nodeindex_t n=itemstarts[i];n<itemstarts[i+1];n++) {
#ifdef CHECK
      if (parents[n]>=itemstarts[i]) {
	warn(0,"1parent[%d]=%d, > %d",n,parents[n],itemstarts[i]);
      }
#endif  
      if (!counters[n])
	continue;
      register int p=parents[n];
      counter_t &nc=ncounters[p];
      if ((!nc) && (p)) {
	//find the item of the parent
	while (p<(int)itemstarts[citem]) {
	  //err(0,"Parent's items are non-ordered. 1 %d %d",n,p);
	  citem--;
	}
	while (p>=(int)itemstarts[citem+1]) citem++;
	nonzerolists[p]=heads[citem];
	heads[citem]=p;
      } 
      nc+=counters[n];
    }
    //fill upwards counters
#ifdef SINGLET
    lastsinglet=i;
#endif
    for (item_t ci=i-1;ci>=0;ci--) {
      size_t incount=0;
      item_t citem=0;
      nullcount+=itemstarts[ci+1]-itemstarts[ci];
      incount=itemstarts[ci+1]-itemstarts[ci];
      for (register nodeindex_t n=heads[ci];n;n=nonzerolists[n]) {
	register int p=parents[n];
	counter_t &nc=ncounters[p];
	if ((!nc)&&(p)) {
	  while (p<(int)itemstarts[citem]) {
	    //err(0,"Parent's items are non-ordered 2 %d %d.",n,p);
	    citem--;
	  }
	  while (p>=(int)itemstarts[citem+1]) citem++;
	  nonzerolists[p]=heads[citem];
	  heads[citem]=p;
	}
	nc+=ncounters[n];
	nitemcounters[ci]+=ncounters[n];
	nullcount--;
	incount--;
	/*	  register int t=nonzerolists[n];
		  nonzerolists[n]=0; //fill with zero after ourselves
		  n=t;*/
      }
      int lposcount=itemstarts[ci+1]-itemstarts[ci]-incount;
      if (nitemcounters[ci]<min_supp) {
	nullcount+=lposcount;
      } 
#ifdef SINGLET
      else {
	if ((lposcount>1))
	  lastsinglet=ci;
      }
#endif
    }//for upper levels
    return nullcount;
  } //cpb_aggregate_sparse



  //npos does not need to be filled with zeroes
  //nncounters does not need to be filled with 0
  //nparents does not need to be filled with zeroes
  void  DINLINE calcproject_dense(item_t i,
				  itemstarts_t nitemstarts,
				  parents_t npos,
				  parents_t nparents,
				  itemstarts_t itemstarts,
				  itemcounters_t nitemcounters,
				  counters_t ncounters,
				  counters_t nncounters,
				  parents_t parents) {
    nitemstarts[0]=1;
    nodeindex_t nn=0;
    npos[0]=0;
    nparents[0]=0;
#ifdef STATS
    projinlen+=itemstarts[i];
#endif
    for (item_t ci=0;ci<i;ci++) {
      nitemstarts[ci]=nn+1;
      for(nodeindex_t n=itemstarts[ci];n<itemstarts[ci+1];n++) {
	if ((nitemcounters[ci]<min_supp)||(!ncounters[n])) {
	  npos[n]=npos[parents[n]];
	} else {
	  npos[n]=++nn;
	  nparents[nn]=npos[parents[n]];
	  nncounters[nn]=ncounters[n];
	}
      }//for nodes on current level
    }//for level (item)
    nn++;
#ifdef STATS
    projtotlen+=nn;
#endif
    nitemstarts[i]=nn;
    //    for (item_t ci=i;ci<=(int)itemcount;ci++)
    //      nitemstarts[ci]=nn;
  }
  


  //npos does not need to be filled with zeroes
  //nncounters does not need to be filled with 0
  //nparents does not need to be filled with zeroes
  void  DINLINE calcproject_sparse(item_t i,
				   itemstarts_t nitemstarts,
				   parents_t npos,
				   parents_t nparents,
				   itemstarts_t itemstarts,
				   itemcounters_t nitemcounters,
				   counters_t ncounters,
				   counters_t nncounters,
				   parents_t parents) {
    nitemstarts[0]=1;
    nodeindex_t nn=0;
    npos[0]=0;
    nparents[0]=0;
#ifdef STATS
    projinlen+=itemstarts[i];
#endif
    for (item_t ci=0;ci<i;ci++) {
      nitemstarts[ci]=nn+1;
      for(register nodeindex_t n=heads[ci];n;) {
	if ((nitemcounters[ci]<min_supp)||(!ncounters[n])) {
	  npos[n]=npos[parents[n]];
	} else {
	  npos[n]=++nn;
	  nparents[nn]=npos[parents[n]];
	  nncounters[nn]=ncounters[n];
	}
	register int t=nonzerolists[n];
	nonzerolists[n]=0; //fill with zero after ourselves
	n=t;	  
      }//for nodes on current level
    }//for level (item)
    nn++;
#ifdef STATS
    projtotlen+=nn;
#endif
    nitemstarts[i]=nn;
    //    for (item_t ci=i;ci<=(int)itemcount;ci++)
    //      nitemstarts[ci]=nn;
  }


  //! global parameters to the recoutput function
  parents_t ro_parents;
  //! global parameters to the recoutput function
  itemstarts_t ro_itemstarts;
  //! global parameters to the recoutput function
  uint32_t ro_supp;
  


  void recoutput(item_t i,nodeindex_t n) {
    output->pushItem(ro_supp,backorder[i]);
    while ((n=ro_parents[n])) {
      while (ro_itemstarts[i]>n)
	i--;
      recoutput(i,n);
    }
    output->popStack();
  }

  void dononprojfp_dense(int recdepth,item_t lastitem,parents_t parents,counters_t counters,itemstarts_t itemstarts,itemcounters_t itemcounters, uint32_t support, int singletlevel) {
#ifdef STATS
    numden++;
    SETLEVEL_WC(profrecdepth,recdepth+1);
#endif
    if (!lastitem)
      return;
    counters_t ncounters=alloccounters.getnextblock(itemstarts[lastitem-1]);
    itemcounters_t nitemcounters=allocitemstarts.getnextblock(lastitem).mem[0];
    for(item_t i=lastitem-1;i>=0;i--) {
      if (itemcounters[i]<min_supp) {
	//info(0,"last item %d, infreq item %d, counter=%d",lastitem,i,itemcounters[i]);
	continue;
      } else if (i<singletlevel) {
#ifdef STATS
	den2sing++;
#endif
	//freqset.push_back(backorder[i]);
	SETLEVEL(profrecdepth,0);	
	ro_parents=parents;
	ro_itemstarts=itemstarts;
	ro_supp=itemcounters[i];
	for (nodeindex_t n=itemstarts[i];n<itemstarts[i+1];n++) {
	  if (counters[n])
	    recoutput(i,n);
	}
	//freqset.pop_back();
	SETLEVEL(profrecdepth,recdepth+1);
	continue;
	} else if (itemcounters[i]==support) {
	//there is a 100% association rule from the current condition
	//to item i. No conditional aggregation is needed.
	//freqset.push_back(backorder[i]);
	SETLEVEL(profrecdepth,0);
#ifndef NOOPTOUT
	output->pushItem(support,backorder[i]);
#else
	output->pushFreqItems(itemcounters[i],freqset);
#endif
	dononprojfp_dense(recdepth+1,i,parents,counters,itemstarts,itemcounters,support, singletlevel);
	//freqset.pop_back();
	SETLEVEL(profrecdepth,recdepth+1);
#ifndef NOOPTOUT
	output->popStack();
#endif
	continue;
	}
      //info(0,"lastitem %d, item %d",lastitem,i);
      //ncounters.resize(itemstarts[i],0);
      FILLZERO(nitemcounters,i+1);
      //      nitemcounters.resize(itemcount,0);
      SETLEVEL_WC(profaggr,recdepth+1);
      size_t nullcount=cpb_aggregate_dense(i,itemstarts,ncounters,counters,parents,nitemcounters);
      nullcount*=100;
      SETLEVEL_WC(profaggr,0);
#ifdef SINGLET
      if (singletlevel>=i) {//remaining of the fp tree is a chain
#ifdef STATS
	den2sing++;
#endif
	//freqset.push_back(backorder[i]);
	SETLEVEL(profrecdepth,0);	
	output->pushItem(itemcounters[i],backorder[i]);
	ro_parents=parents;
	ro_itemstarts=itemstarts;
	for (item_t ci=i-1;ci>=0;ci--) {
	  for (nodeindex_t n=itemstarts[ci];n<itemstarts[ci+1];n++) {
	    if ((ro_supp=ncounters[n])>=min_supp) {
	      recoutput(ci,n);
	    }
	  }
	}
	//freqset.pop_back();
	SETLEVEL(profrecdepth,recdepth+1);
	output->popStack();
      } else
#endif
      if (((int)recdepth<projlevel) || (nullcount>=(itemstarts[i]*projpercent))) { //if more than projpercent % gain on projection
#ifdef STATS
	den2proj++;
#endif
	//--------- PROJECT ----------
	//select surviving nodes and calculate new offsets
	itemstarts_t nitemstarts=allocitemstarts.getnextblock(i+1).mem[0];
	FILLZERO(nitemstarts,i+1);
	//      nitemstarts.resize(itemcount+1);
	projallocator_t::memblock_t nm=allocprojections.getnextblock(itemstarts[i]);
	parents_t nparents=nm.mem[0];
	//nparents.resize(itemstarts[i]);
	//FILLZERO(nparents,itemstarts[i]);NO NEED
	counters_t nncounters=nm.mem[1];
	//nncounters.resize(itemstarts[i]);
	//FILLZERO(nncounters,itemstarts[i]);NO NEED
	//	parents_t npos;
	//npos.resize(itemstarts[i]);
	//FILLZERO(nposit,itemstarts[i]);NO NEED
	SETLEVEL_WC(profproj,recdepth+1);
	calcproject_dense(i,nitemstarts,nposit,nparents,itemstarts,nitemcounters,ncounters,nncounters,parents);
	SETLEVEL_WC(profproj,0);
	//cleanup
	//do recursive call
	//freqset.push_back(backorder[i]);
	SETLEVEL(profrecdepth,0);
#ifndef NOOPTOUT
	output->pushItem(itemcounters[i],backorder[i]);
#else
	output->pushFreqItems(itemcounters[i],freqset);
#endif
	dononprojfp_dense(recdepth+1,i,nparents,nncounters,nitemstarts,nitemcounters,itemcounters[i],lastsinglet);
	//freqset.pop_back();
	SETLEVEL(profrecdepth,recdepth+1);
#ifndef NOOPTOUT
	output->popStack();
#endif
	allocprojections.freelastblock();
	allocitemstarts.freelastblock();
      } else {
	//freqset.push_back(backorder[i]);
	SETLEVEL(profrecdepth,0);	
#ifndef NOOPTOUT
	output->pushItem(itemcounters[i],backorder[i]);
#else
	output->pushFreqItems(itemcounters[i],freqset);
#endif
	if (((int)itemstarts[i]>densesize)&&(nullcount>=(itemstarts[i]*densepercent))) {
#ifdef STATS
	  den2sp++;
#endif
	  dononprojfp_sparse(recdepth+1,i,parents,ncounters,itemstarts,nitemcounters,itemcounters[i],lastsinglet);	  
	} else {
	  dononprojfp_dense(recdepth+1,i,parents,ncounters,itemstarts,nitemcounters,itemcounters[i],lastsinglet);	  
	}
	//freqset.pop_back();
	SETLEVEL(profrecdepth,recdepth+1);
#ifndef NOOPTOUT
	output->popStack();
#endif
      }
      // **** substitute with FILLZERO
      //cpb_clear(i,ncounters,parents,itemstarts);
      FILLZERO(ncounters,itemstarts[i]);
    }//for item to add to the current condition
    alloccounters.freelastblock();
    allocitemstarts.freelastblock();
    SETLEVEL(profrecdepth,recdepth);
  }//dononprojfp_dense

  void dononprojfp_sparse(int recdepth,item_t lastitem,parents_t parents,counters_t counters,itemstarts_t itemstarts,itemcounters_t itemcounters,uint32_t support, int singletlevel) {
#ifdef STATS
    numsp++;
#endif
    if (!lastitem)
      return;
    SETLEVEL_WC(profrecdepth,recdepth+1);
    counters_t ncounters=alloccounters.getnextblock(itemstarts[lastitem-1]);
    itemcounters_t nitemcounters=allocitemstarts.getnextblock(lastitem).mem[0];
    for(item_t i=lastitem-1;i>=0;i--) {
      if (itemcounters[i]<min_supp) {
	//info(0,"last item %d, infreq item %d, counter=%d",lastitem,i,itemcounters[i]);
	continue;
      } else if (itemcounters[i]==support) {
	//there is a 100% association rule from the current condition
	//to item i. No conditional aggregation is needed.
	//freqset.push_back(backorder[i]);
	SETLEVEL(profrecdepth,0);
#ifndef NOOPTOUT
	output->pushItem(support,backorder[i]);
#else
	output->pushFreqItems(itemcounters[i],freqset);
#endif
	dononprojfp_sparse(recdepth+1,i,parents,counters,itemstarts,itemcounters,support,singletlevel);
	//freqset.pop_back();
	SETLEVEL(profrecdepth,recdepth+1);
#ifndef NOOPTOUT
	output->popStack();
#endif
	continue;
      }
      FILLZERO(nitemcounters,i+1);
      SETLEVEL_WC(profaggr,recdepth+1);
      size_t nullcount=cpb_aggregate_sparse(i,itemstarts,ncounters,counters,parents,nitemcounters);
      nullcount*=100;
      SETLEVEL_WC(profaggr,0);
#ifdef SINGLET
      if (lastsinglet>=i) {//remaining of the fp tree is a chain
#ifdef STATS
	sp2sing++;
#endif
	//freqset.push_back(backorder[i]);
	SETLEVEL(profrecdepth,0);	
	output->pushItem(itemcounters[i],backorder[i]);
	ro_parents=parents;
	ro_itemstarts=itemstarts;
	for (item_t ci=i-1;ci>=0;ci--) {
	  for (register nodeindex_t n=heads[ci];n;n=nonzerolists[n]) {
	    if ((ro_supp=ncounters[n])>=min_supp) {
	      recoutput(ci,n);
	    }
	  }
	}
	//freqset.pop_back();
	SETLEVEL(profrecdepth,recdepth+1);
	output->popStack();
      } else
#endif
      if (((int)recdepth<projlevel) || (nullcount>=(itemstarts[i]*projpercent))) { //if more than projpercent % gain on projection
#ifdef STATS
	sp2proj++;
#endif
	//--------- PROJECT ----------
	//select surviving nodes and calculate new offsets
	itemstarts_t nitemstarts=allocitemstarts.getnextblock(i+1).mem[0];
	FILLZERO(nitemstarts,i+1);
	//      nitemstarts.resize(itemcount+1);
	projallocator_t::memblock_t nm=allocprojections.getnextblock(itemstarts[i]);
	parents_t nparents=nm.mem[0];
	//nparents.resize(itemstarts[i]);
	//FILLZERO(nparents,itemstarts[i]);NO NEED
	counters_t nncounters=nm.mem[1];
	//nncounters.resize(itemstarts[i]);
	//FILLZERO(nncounters,itemstarts[i]);NO NEED
	//	parents_t npos;
	//npos.resize(itemstarts[i]);
	//FILLZERO(nposit,itemstarts[i]);NO NEED
	SETLEVEL_WC(profproj,recdepth+1);
	calcproject_sparse(i,nitemstarts,nposit,nparents,itemstarts,nitemcounters,ncounters,nncounters,parents);
	SETLEVEL_WC(profproj,0);
	//cleanup
	//do recursive call
	//freqset.push_back(backorder[i]);
	SETLEVEL(profrecdepth,0);
#ifndef NOOPTOUT
	output->pushItem(itemcounters[i],backorder[i]);
#else
	output->pushFreqItems(itemcounters[i],freqset);
#endif
	dononprojfp_sparse(recdepth+1,i,nparents,nncounters,nitemstarts,nitemcounters,itemcounters[i],lastsinglet);
	//freqset.pop_back();
	SETLEVEL(profrecdepth,recdepth+1);
#ifndef NOOPTOUT
	output->popStack();
#endif
	allocprojections.freelastblock();
	allocitemstarts.freelastblock();
      } else {
	//return nonzerolists to zero state
	for (item_t ci=i-1;ci>=0;ci--) {
	  for (register nodeindex_t n=heads[ci];n;) {
	    register int t=nonzerolists[n];
	    nonzerolists[n]=0; //fill with zero after ourselves
	    n=t;
	  }
	}
	//freqset.push_back(backorder[i]);
	SETLEVEL(profrecdepth,0);	
#ifndef NOOPTOUT
	output->pushItem(itemcounters[i],backorder[i]);
#else
	output->pushFreqItems(itemcounters[i],freqset);
#endif
	if (((int)itemstarts[i]>densesize)&&(nullcount>=(itemstarts[i]*densepercent))) {
	  dononprojfp_sparse(recdepth+1,i,parents,ncounters,itemstarts,nitemcounters,itemcounters[i],lastsinglet);	  
	} else {
#ifdef STATS
	  sp2den++;
#endif
	  dononprojfp_dense(recdepth+1,i,parents,ncounters,itemstarts,nitemcounters,itemcounters[i],lastsinglet);	  
	}
	//freqset.pop_back();
	SETLEVEL(profrecdepth,recdepth+1);
#ifndef NOOPTOUT
	output->popStack();
#endif
      }	
      cpb_clear(i,ncounters,parents,itemstarts);
    }//for item to add to the current condition
    alloccounters.freelastblock();
    allocitemstarts.freelastblock();
    SETLEVEL(profrecdepth,recdepth);
}//donoprojfp_sparse


};//class NonOrdFp


#ifdef FIMI
//auxiliary routines for I/O 


void FIMIcoord::applyConfig(string filename, string infilename) {
 

  bufsize=DEFAULT_BUFSIZE;


  info(0,"Using output file %s.",filename.c_str());
  if (filename.size()) {
    fdout=open(filename.c_str(),O_WRONLY|O_CREAT|O_TRUNC,S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH);
    //fout = fopen(filename.c_str(),"w");
    if //(!fout) {
      (fdout<0){
      err(0,"FrequISet: error opening output file: %s",strerror(errno));
    }
#ifdef NOOPTOUT
    fout=fdopen(fdout,"w");
    //    setbuffer(fout,outbuf,bufsize);
#endif
    outbuf = new char[bufsize];
    bufend=outbuf+bufsize;
    buffree=outbuf;
#ifndef DBGG
#else
    //setvbuf(fout,NULL,_IONBF,0);
#endif
  }
  else {
    fout=NULL;
    fdout=0;
    //outbuf=new char[bufsize];
  }

  fin = fopen(infilename.c_str(),"r");
  if (!fin) {
    err(0,"FrequISet: error opening input file: %s",strerror(errno));
  }
  inbuf = new char[bufsize];
  setbuffer(fin,inbuf,bufsize);
}

inline void FIMIcoord::pushAllRows() {
  item_vector_t Items;

  fseek(fin,SEEK_SET,0);
  char *inbuf = new char[bufsize+1];
  char *bufp,*bufend;
  char c;
  int r=fread(inbuf,1,bufsize,fin);
  bufp=inbuf; bufend=inbuf+r;
  bool ateof=false;
  while(!ateof) {
    Items.clear();

    do {
      int item=0, pos=0;
      c=*bufp;
      if ((++bufp)==bufend) {
	bufp=inbuf;
	if (feof(fin)) {
	  ateof=true;
	  *bufp='\n';
	} else {
	  bufend=bufp+fread(inbuf,1,bufsize,fin);
	}
      }
      while((c >= '0') && (c <= '9')) {
	item *=10;
	item += int(c)-int('0');
	//	c = getc(fin);

	c=*bufp;

	if ((++bufp)==bufend) {
	  bufp=inbuf;
	  if (feof(fin)) {
	    ateof=true;
	    *bufp='\n';
	  } else {
	    bufend=bufp+fread(inbuf,1,bufsize,fin);
	  }
	}
	pos++;
      }
      if(pos) Items.push_back(item);
      if (item>maxitem)
	maxitem=item;
    }while((c != '\n') && (!ateof));

    if (Items.size()) {
      algo->In(Items);
    }
  }
  if (!endposst) {
    endposst=new int[maxitem+1];
    endpossp=0;
    endposst[endpossp]=0;
    //item render stack
    itemtt=new char[(maxitem+1)*8];
    itemtt[0]=1;
    itemtt[1]='0';
    histo=new int[maxitem+1];
    for (int i=0;i<=maxitem;i++)
      histo[i]=0;
  }
  delete [] inbuf;
}

void FIMIcoord::doCoding() {
  numfreq=0;
  dispTotalTime();
  status(2,"Initializing algorithm.");
  
  //algo->init(NULL,this,0);//hope it does not need the link

  linebuf=new char[LINEBUFLEN];
  endposst=NULL;
  //  endposst.push(0);
  maxitem=0;
  int pass=1;
  do {
    dispTotalTime();
    status(2,"Starting pass %d",pass++);
    algo->startPass();
    dispTotalTime();
    pushAllRows();
    status(2,"End of data reached.");
    dispTotalTime();
  } while (algo->finishPass());
  dispTotalTime();
  algo->finish();
  dispTotalTime();
  status(2,"closing file");
  info(0,"%lld frequent itemsets returned.",numfreq);
  if (fdout) {
    bufdflush();
    delete [] outbuf;
#ifdef NOOPTOUT
    fclose(fout);
#else
    close(fdout);
#endif
  }
  fclose(fin);
  delete [] inbuf;
  dispHisto();
}



void FIMIcoord::pushFreqItems(size_t support, item_vector_t &items) {
  if (fdout) {
    char b[1000];
    int len;
    for(unsigned int i=0; i<items.size(); i++) {
      sprintf(b, "%d %n", items[i],&len);
      bufdwrite(b,len);
    }
    sprintf(b, "(%d)\n%n", support,&len);
    bufdwrite(b,len);    
  }
  numfreq++;
}

void __attribute__((always_inline)) renderi(char *ip,item_t item) {
  int pos=7;
  while (item) {
    ip[pos--]='0'+item%10;
    item/=10;
  }
  pos++;
  int pos2=1;
  while (pos<=7) {
    ip[pos2++]=ip[pos++];
  }
  ip[0]=pos2-1;
}


void __attribute__((always_inline)) FIMIcoord::putend(char *buf,item_t item,int &nextt,size_t  support,char **last) {
#define BLEN 100
  static char r[BLEN];
  static int lastsupp=-1;
  static int lastsupppos=0;
  int pos=BLEN-1;
  nextt=0;
  char *ip=itemtt+(item<<3);
  if (! *ip) {
    renderi(ip,item);
  }
  int len=ip[0];
  nextt+=len;
  pos=1;
  while(len) {
    *(buf++)=ip[pos++];
    len--;
  }
  *(buf++)=' ';
  nextt++;
  *(buf++)='(';
  pos=BLEN-1;
  if (!support) {
    *(buf++)='0';
  } else {
    if (support !=(size_t)lastsupp) {
      while (support) {
	r[pos--]='0'+support%10;
	support/=10;
      }
      pos++;
    } else {
      pos=lastsupppos;
    }
    lastsupppos=pos;
    lastsupp=support;
    int len=BLEN-pos;
    while(len) {
      *(buf++)=r[pos++];
      len--;
    }
    //    memcpy(buf,r+pos,len);
    //buf+=len;
  }
  *(buf++)=')';
  *(buf++)='\n';
  *last=buf;
  *(buf++)=0;
  }

void __attribute__((always_inline)) FIMIcoord::bufdwrite(char *buf,size_t len) {
  if ((int)len>bufend-buffree) {
    int part=bufend-buffree;
    mempcpy(buffree,buf,part);
    buffree=outbuf;
    while (buffree<bufend) {
      int ret=write(fdout,buffree,bufend-buffree);
      if (ret<0) {
	perror("write");
      } else
	buffree+=ret;
    }
    buffree=(char*)mempcpy(outbuf,buf+part,len-part);
  } else {
    buffree=(char*)mempcpy(buffree,buf,len);
  }
}

void FIMIcoord::bufdflush() {
  bufend=buffree;
  buffree=outbuf;
  while (buffree<bufend) {
    int ret=write(fdout,buffree,bufend-buffree);
    if (ret<0) {
      perror("write");
    } else
      buffree+=ret;
  }
  bufend=outbuf+bufsize;
  buffree=outbuf;
}






int main(int argc, char *argv[]) {
  NonOrdFp algo;
  FIMIcoord crd;
  
  if (argc<3) {
    fprintf(stderr,"Usage: %s database min_supp [outputfile [configfile]]\n",argv[0]);
    fprintf(stderr,"configfile, if specified, may contain the following lines:\n");
    fprintf(stderr,"simultprojthres=<number>\n\tif the fp tree's size is above this threshold, then use simultaneous projection on the first level.\n"
"projpercent=<0-101>\n\tif the number of deleted FP tree nodes exceeds"
" this percentage, then projection is calculated.\n"
"densepercent=<0-101>\n\tif the number of deleted FP tree nodes exceeds"
" this percentage, then sparse aggregation is used.\n"
"projlevel=<number>\n\tall the projections until this level are done.\n"
"filterdupl=<number>\n\tif non-zero, filter duplicate items from transactions.\n"
	    "\t(projlevel 0 and projpercent 101 means do not do any projections)\n");
    exit(1);
  }
  if (argc==3) {
    crd.applyConfig("",argv[1]);
  } else {
    crd.applyConfig(argv[3],argv[1]);
  }
  algo.min_supp=atoi(argv[2]);
  algo.output=&crd;
  crd.algo=&algo;

  if (argc>4) {
    configuration root;
    ifstream fin(argv[4]);
    root.appendFromFile(fin);
    fin.close();
    algo.custApplyConfig(&root);
  } else {
    configuration root;
    algo.custApplyConfig(&root); //leave everything as the default
  }

  crd.doCoding();

  return 0;
}


#endif

#ifndef FIMI
REGISTER_CLASS( nonordfp, NonOrdFp );
#endif
